/**
 * Supporting code for "Computing closed essential surfaces in knot
 * complements", B.A. Burton, A. Coward and S. Tillmann, arXiv:1212.1531.
 *
 * Copyright (c) 2012, Benjamin A. Burton.
 *
 * Released under the GNU General Public License, version 2 or greater.
 *
 * This is a wrapper program that identifies candidate essential
 * surfaces by enumerating all admissible extreme rays of the cone Q_0(T).
 *
 * - The bulk of the code is in the header unified.hpp, which is
 *   thoroughly documented.
 *
 * - This short wrapper is not documented, but will be in a future version
 *   of this software bundle.
 *
 * This code needs to be built against Regina 4.94 or later.  Until
 * Regina 4.94 is released, you can build it against the current trunk in
 * Regina's online source code repository, which you can check out from
 * http://regina.sourceforge.net/ .
 *
 * Please check again for future versions of this software bundle at
 * http://www.maths.uq.edu.au/~bab/code/ .
 *
 * - Ben Burton, 15 December 2012.
 */

#include "unified.hpp"
#include "file/nxmlfile.h"
#include <ctime>

/**
 * Output information on how to use this program, and then terminate the
 * entire program with return value 1.
 *
 * @param program the name of this program (typically argv[0]).
 */
void usage(const char* program) {
    std::cerr << "Usage:\n\n"
        << program << " <name> <isosig>\n"
        "\n";
    ::exit(1);
}

template <typename LPConstraint>
bool buildAndVerify(const TreeEnumeration<LPConstraint>& tree,
        void* matchingEqns) {
    std::cout << "SOLN #" << tree.nSolns() << ": ";
    NNormalSurface* f = tree.buildSurface();
    bool ok = tree.verify(f, *static_cast<NMatrixInt*>(matchingEqns));
    if (! ok)
        std::cout << "BROKEN - " << f->toString() << std::endl;
    else
        std::cout << "Ok, chi = " << f->getEulerCharacteristic() << std::endl;
    delete f;
    return true;
}

class AdHocList : public regina::NNormalSurfaceList {
    private:
        regina::NMatrixInt* eqns_;
        bool broken_;

    public:
        AdHocList(NTriangulation* tri) : NNormalSurfaceList(),
                eqns_(regina::makeMatchingEquations(tri, QUAD)),
                broken_(false) {
            flavour = QUAD;
            embedded = true;

            setPacketLabel("Surfaces");
            tri->insertChildLast(this);
        }

        template <typename LPConstraint>
        void push(NNormalSurface* s,
                const TreeEnumeration<LPConstraint>& tree) {
            surfaces.push_back(s);

            if (! tree.verify(s, eqns_))
                broken_ = true;
        }

        bool broken() const {
            return broken_;
        }
};

template <typename LPConstraint>
bool pushToList(const TreeEnumeration<LPConstraint>& tree, void* list) {
    NNormalSurface* f = tree.buildSurface();
    static_cast<AdHocList*>(list)->push(f, tree);
    return true;
}

/**
 * The main body of the program.
 *
 * For each isomorphism signature passed on the command line, we build
 * the corresponding triangulation and then run away to perform the tree
 * traversal algorithm.
 */
int main(int argc, char* argv[]) {
    char mode = 0;

    if (argc != 3)
        usage(argv[0]);

    const char* name = argv[1];
    const char* isoSig = argv[2];

    // Process the next triangulation on the command line:
    NTriangulation* t = NTriangulation::fromIsoSig(isoSig);
    t->setPacketLabel(name);
    if (! t) {
        std::cerr << name << " - ERROR: Could not decode isosig: " << isoSig
            << std::endl;
        return 1;
    }

    if (t->isOrientable())
        t->orient();

    clock_t uTime = clock();

    TreeEnumeration<LPConstraintNonSpun> search(t, NNormalSurfaceList::QUAD);
    if (search.constraintsBroken()) {
        std::cerr << name << " - ERROR: Constraints broken." << std::endl;
        delete t;
        return 1;
    }

    AdHocList* list = new AdHocList(t);
    search.run(&pushToList, list);

    uTime = clock() - uTime;

    std::cout << name << ' ' << search.nSolns() << ' '
        << search.nVisited() << ' '
        << (list->broken() ? "ERROR" : "ok") << ' '
        << uTime << std::endl;

    regina::writeXMLFile((std::string(name) + ".rga").c_str(), t);

    delete t;
    return 0;
}
