/**
 * Supporting code for "Computing closed essential surfaces in knot
 * complements", B.A. Burton, A. Coward and S. Tillmann, arXiv:1212.1531.
 *
 * Copyright (c) 2012, Benjamin A. Burton.
 *
 * Released under the GNU General Public License, version 2 or greater.
 *
 * This is a wrapper program that tests whether a given surface is
 * incompressible:
 *
 * - The bulk of the code is in the header unified.hpp, which is
 *   thoroughly documented.
 *
 * - This short wrapper is not documented, but will be in a future version
 *   of this software bundle.
 *
 * This code needs to be built against Regina 4.94 or later.  Until
 * Regina 4.94 is released, you can build it against the current trunk in
 * Regina's online source code repository, which you can check out from
 * http://regina.sourceforge.net/ .
 *
 * Please check again for future versions of this software bundle at
 * http://www.maths.uq.edu.au/~bab/code/ .
 *
 * - Ben Burton, 15 December 2012.
 */

#include "unified.hpp"
#include <file/nxmlfile.h>
#include <ctime>
#include <cstdlib>

NTriangulation* findPiece(NTriangulation* t, long ec, bool ideal) {
    unsigned nComp = t->splitIntoComponents();
    NTriangulation* c = static_cast<NTriangulation*>(t->getFirstTreeChild());
    bool found = false;
    while (c) {
        if (ideal) {
            if (c->getNumberOfBoundaryComponents() == 2 &&
                    ((c->getBoundaryComponent(0)->
                        getEulerCharacteristic() == 0 &&
                      c->getBoundaryComponent(1)->
                        getEulerCharacteristic() == ec) ||
                     (c->getBoundaryComponent(0)->
                        getEulerCharacteristic() == ec &&
                      c->getBoundaryComponent(1)->
                        getEulerCharacteristic() == 0)))
                found = true;
        } else {
            if (c->getNumberOfBoundaryComponents() == 1 &&
                    c->getBoundaryComponent(0)->getEulerCharacteristic() == ec)
                found = true;
        }

        if (found) {
            c->makeOrphan();
            delete t;
            return c;
        }

        c = static_cast<NTriangulation*>(c->getNextTreeSibling());
    }
    delete t;
    return 0;
}

// Side-effect: deletes t.
bool compresses(NTriangulation* t, bool ideal) {
    t->intelligentSimplify();

    if (ideal) {
        t->idealToFinite();
        t->intelligentSimplify();
    }

    // These can be large triangulations.
    // Barycentric subdivide and try again to simplify, just in case
    // this makes things even a little smaller.
    {
        NTriangulation* u = new NTriangulation(*t);
        u->barycentricSubdivision();
        u->intelligentSimplify();
        u->intelligentSimplify();
        if (u->getNumberOfTetrahedra() < t->getNumberOfTetrahedra()) {
            delete t;
            t = u;
        } else
            delete u;
    }

    long ec = t->getBoundaryComponent(0)->getEulerCharacteristic();
    if (ideal && ec == 0)
        ec = t->getBoundaryComponent(1)->getEulerCharacteristic();

    NNormalSurface* ans;
    NMatrixInt* eqns;
    NTriangulation* crush;
    while (true) {
        if (t->getNumberOfVertices() != t->getNumberOfBoundaryComponents()) {
            std::cout << "ERROR: Too many vertices" << std::endl;
            std::exit(9);
        }
        std::cout << "Searching for disc or sphere ... "; std::cout.flush();
        TreeSingleSoln<LPConstraintEuler, BanTorusBoundary> search(t,
            NNormalSurfaceList::STANDARD);
        if (search.find()) {
            std::cout << "found ... "; std::cout.flush();
            ans = search.buildSurface();
            if (! search.verify(ans)) {
                std::cout << "ERROR: Broken surface" << std::endl;
                writeXMLFile("broken.rga", t);
                std::exit(6);
            }
            if (! (ans->isConnected())) {
                std::cout << "ERROR: Disconnected surface" << std::endl;
                std::exit(7);
            }
            if (ans->getEulerCharacteristic() <= 0) {
                std::cout << "ERROR: Surface with genus" << std::endl;
                std::exit(8);
            }

            crush = ans->crush();
            std::cout << "crushing ... "; std::cout.flush();
            t = findPiece(crush, ec, ideal);
            if (! t) {
                // We compressed.
                std::cout << "compressed." << std::endl;
                return true;
            }
            std::cout << "simplified." << std::endl;
            // We simplified the triangulation.  Around we go again.
        } else {
            std::cout << "not found." << std::endl;
            delete t;
            return false;
        }
    }
}

bool incompressible(const NNormalSurface* f) {
    if (f->isThinEdgeLink().first) {
        std::cout << "Thin edge link" << std::endl;
        return false;
    }

    NTriangulation* cut = f->cutAlong();
    cut->intelligentSimplify();

    long nComp = cut->splitIntoComponents();
    if (nComp != 2) {
        std::cout << "ERROR: Not two components" << std::endl;
        std::exit(5);
    }

    NTriangulation* t1 = static_cast<NTriangulation*>(cut->getFirstTreeChild());
    NTriangulation* t2 = static_cast<NTriangulation*>(t1->getNextTreeSibling());

    bool result;
    if (t1->isIdeal()) {
        if (t1->getNumberOfTetrahedra() < t2->getNumberOfTetrahedra())
            result = compresses(t1, true) || compresses(t2, false);
        else
            result = compresses(t2, false) || compresses(t1, true);
    } else {
        if (t1->getNumberOfTetrahedra() < t2->getNumberOfTetrahedra())
            result = compresses(t1, false) || compresses(t2, true);
        else
            result = compresses(t2, true) || compresses(t1, false);
    }

    delete cut;
    return ! result;
}

struct ID {
    long index;
    long ec;

    inline bool operator < (const ID& rhs) const {
        return (ec > rhs.ec || (ec == rhs.ec && index < rhs.index));
    }
};

int main(int argc, char* argv[]) {
    if (argc != 2)
        return 1;

    NPacket* tree = readXMLFile(argv[1]);
    if (! tree)
        return 2;

    std::cout << "Triangulation: " << tree->getPacketLabel() << std::endl;

    NNormalSurfaceList* s = dynamic_cast<NNormalSurfaceList*>(
        tree->getFirstTreeChild());
    if (! s)
        return 3;

    std::cout << "Surfaces: " << s->getNumberOfSurfaces() << std::endl;

    clock_t uTime = clock();

    ID* id = new ID[s->getNumberOfSurfaces()];
    unsigned i;
    for (i = 0; i < s->getNumberOfSurfaces(); ++i) {
        id[i].index = i;
        id[i].ec = s->getSurface(i)->getEulerCharacteristic().longValue();
    }
    std::sort(id, id + s->getNumberOfSurfaces());

    bool large = false;
    const NNormalSurface* f;
    NLargeInteger ec;
    for (unsigned i = 0; i < s->getNumberOfSurfaces(); ++i) {
        f = s->getSurface(id[i].index);
        ec = f->getEulerCharacteristic();
        std::cout << id[i].index << ": chi = " << f->getEulerCharacteristic()
            << std::endl;
        if (ec == 0) {
            std::cout << "ABORT!" << std::endl;
            return 4;
        }
        if (incompressible(f)) {
            std::cout << "Incompressible" << std::endl;
            std::cout << "Result: Large" << std::endl;
            large = true;
            break;
        }
    }
    if (! large)
        std::cout << "Result: Small" << std::endl;

    delete[] id;

    uTime = clock() - uTime;
    std::cout << "Time: " << uTime << std::endl;

    delete tree;
    return 0;
}
